/***************************************************************************
                          sglwindow.cpp  -  description
                             -------------------
    begin                : Wed Jan 23 2002
    copyright            : (C) 2002 by Andre Stein
    email                : andre_stein@web.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "sglwindow.h"
#include <SDL.h>
#include <stdlib.h>

/* LIGHTING CONSTs */
const float WHOLE_SCENE_LIGHT[] = { 0.25f, 0.25f, 0.25f, 1.0f};
const float DIFFUSE_LIGHT[]  = { 0.6f, 0.6f, 0.6f, 1.0f};
const float SPECULAR_LIGHT[] = {-.1f, -0.1f, -0.1f, 1.0f};
const float LIGHT_POSITION1[] = { 0.0f, -5.0f,0.0f, 1.0f};
const float LIGHT_POSITION2[] = { 0.0f, 5.0f,0.0f, 1.0f};


const float SPOT_DIRECTION[] = {0.0f,0.0f,-1.0f};

const float MATERIAL[] = {1.0f,1.0f,1.0f,1.0f};


SGLWindow::SGLWindow()
{
        gameState = STATE_LOGO;
}
SGLWindow::~SGLWindow()
{
        /**
          Cleanup
         **/

        //Save highscores
        highscoreList.save(HIGHSCORE_FILENAME);
}
/** No descriptions */
void SGLWindow::init(SConfiguration* config, SDL_Surface* screen)
{
        //copy config file address
        configFile = config;
        //copy main screen
        mainSurface = screen;

        //set caption
        SDL_WM_SetCaption("The Christmas Tetris","");

	glClearColor(0.0f,0.0f,0.0f,0.0f);
	glShadeModel(GL_SMOOTH);
	glClearDepth(1.0f);
	glDepthFunc(GL_LEQUAL);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

        //important state changes
        glEnable(GL_TEXTURE_2D);
      	glEnable(GL_DEPTH_TEST);
        glEnable(GL_CULL_FACE);

        //blend function
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        
        /**
         * LIGHT
         *********/

        glLightModelfv(GL_LIGHT_MODEL_AMBIENT,WHOLE_SCENE_LIGHT);

	//Main
	glLightfv(GL_LIGHT0,GL_AMBIENT,SPECULAR_LIGHT);
	glLightfv(GL_LIGHT0,GL_SPECULAR,SPECULAR_LIGHT);
	glLightfv(GL_LIGHT0,GL_DIFFUSE,DIFFUSE_LIGHT);
	glLightfv(GL_LIGHT0,GL_POSITION,LIGHT_POSITION1);

	glLightfv(GL_LIGHT1,GL_AMBIENT,SPECULAR_LIGHT);
	glLightfv(GL_LIGHT1,GL_SPECULAR,SPECULAR_LIGHT);
	glLightfv(GL_LIGHT1,GL_DIFFUSE,DIFFUSE_LIGHT);
	glLightfv(GL_LIGHT1,GL_POSITION,LIGHT_POSITION2);

	//Spotlight
	glLightf(GL_LIGHT0, GL_SPOT_CUTOFF, 10.0f); //60 Spotlight
	glLightf(GL_LIGHT0,GL_SPOT_EXPONENT, 60.0f);
	glLightfv(GL_LIGHT0,GL_SPOT_DIRECTION,SPOT_DIRECTION);

    	glColorMaterial(GL_FRONT,GL_AMBIENT_AND_DIFFUSE);

	//Materials
	glMaterialfv(GL_FRONT,GL_SPECULAR,MATERIAL);
	glMaterialf(GL_FRONT,GL_SHININESS,32.0f);

	glEnable(GL_COLOR_MATERIAL);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHT1);

        /**
         * INIT OBJECTS
         ****************/

        /** compile, load objects */
        SBox::build();        
         
        //sound system
        soundSystem.init();

        steinsoftLogo.init("data/logo.bmp", SCALE_FADE);
        presentsLogo.init("data/presents.bmp", NORMAL_FADE);
        gameLogo.init("data/game_logo.bmp", NORMAL_FADE,0.001f);
        
        snowFlakes.init();
        snowFlakes.create(); //create first set of flakes

        //gamefield
        gamefield.init(configFile);

        //end scene
        endScene.init();
        
        //create font
        font.init();

        //menu!
        menu.create(&font, configFile);

        //input dialog
        inputDialog.init(&font);

        //credits
        credits.init(&font);

        /**
         * CONFIGURATION
         *****************/

        //Load highscores
        highscoreList.load(HIGHSCORE_FILENAME); 
        //..and set font
        highscoreList.init(&font);
        
        /**
         * MUSIC ON
         ***************/

        //if enabled, play music 
        if (configFile->isMusic())
                soundSystem.playSample(SAMPLE_BACKGROUND);
}
/** No descriptions */
void SGLWindow::resize(int width, int height)
{
	glViewport(0,0,width,height);

	if (height == 0)
		height = 1;	

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	gluPerspective(45.0f,width/height,0.1f,100.0f);

        /**
         * SDL Texture correction
         *****/

        glMatrixMode(GL_TEXTURE);
        glRotatef(180.0f,0.0f,0.0f,1.0f);
        glScalef(-1.0f,1.0f,1.0f);

        /** Init Modelview **/
        
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();  
}

/** No descriptions */
void SGLWindow::update(float frametime)
{
        //for text generation
        static char buffer[256];
        
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glLoadIdentity();

        /**
         * FPS
         **/

        /* float fps = 1000.0f/frametime;
         //draw
         sprintf(buffer,"%f",fps);
         font.printAt(0,0,buffer);*/


        /****
         * HANDLE GAME STATES
         ***********/ 

        if (gameState == STATE_LOGO)
        {
                glTranslatef(0.0f,0.0f,-5.0f);

                steinsoftLogo.update(frametime);

                if (steinsoftLogo.isFinished())
                        gameState = STATE_PRESENTS;
        }
        else if (gameState == STATE_PRESENTS)
        {
                glTranslatef(0.0f,0.0f,-5.0f);

                presentsLogo.update(frametime);

                if (presentsLogo.isFinished())
                        gameState = STATE_INTRO;
        }
        else if (gameState == STATE_INTRO)
        {
                glTranslatef(0.0f,0.0f,-5.0f);
                
                snowFlakes.update(frametime);

                glPushMatrix();
                //game logo
                glScalef(1.0f,2.0f,1.0f);
                gameLogo.update(frametime);

                glPopMatrix();

                if (gameLogo.isFinished())
                {
                        /*gameState = STATE_GAME;
                        gamefield.newBrick();*/
                        gameState = STATE_MENU;
                }
        }
        else if (gameState == STATE_MENU || gameState == STATE_OPTIONS)
        {
                glTranslatef(0.0f,0.0f,-5.0f);

                menu.draw();
                
                snowFlakes.update(frametime);

        }
        else if (gameState == STATE_GAME)
        {
                /** Wanna see flakes if enabled! **/

                if (configFile->isSnowFlakes())
                        snowFlakes.update(frametime);
                
                /** The game! **/
                
                gamefield.update(frametime);

                /**
                 * draw stats (or paused sign)
                 ****************/
                glColor3f(1.0f,0.0f,0.0f);

                if (gamefield.isPaused())
                {
                        font.printAt(430,410,"Paused");        
                }
                else
                {
                        //score
                        sprintf(buffer, "Score: %i", gamefield.getScore());
                        font.printAt(430,450,buffer);
                        //level
                        sprintf(buffer, "Level: %i", gamefield.getLevel());
                        font.printAt(430,430, buffer);
                        //removed lines
                        sprintf(buffer, "Removed Lines: %i", gamefield.getRemovedLines());
                        font.printAt(430,410, buffer);

                        /**
                         * next brick*/

                        font.printAt(430,390, "Next brick:");
                        SBrick brick = gamefield.getNextBrick();

                        glTranslatef(10.0f,7.5f,-30.0f);
                        brick.draw();
                }
                
                /**
                 * how to use message  */

                 glColor3f(1.0f,1.0f,1.0f);
                 font.printAt(495,40,"P",true);
                 font.printAt(550,40,"Pause");
                 font.printAt(495,25,"ESC",true);
                 font.printAt(550,25,"Quit");
                 font.printAt(495,10,"Space", true);
                 font.printAt(550,10,"New Game");
                 

                if (gamefield.isGameOver())
                {
                        /* First check whether we have a new highscore */
                        if (highscoreList.isHighscore(gamefield.getScore()) != 0)
                        {
                                inputDialog.create(MAX_NAME_LENGTH,"Enter your name:");
                                gameState = STATE_INPUT;
                        }
                        //if no highscore, show a little message
                        //to our frustrated user..
                        else
                        {
                                gameState = STATE_NOHIGHSCORE;
                        }
                }
                
        }
        else if (gameState == STATE_END)
        {
                /** The End Scene **/

                glTranslatef(0.0f,0.0f,-5.0f);
                
                //snowFlakes.update(frametime);
                endScene.update(frametime);
        }
        else if (gameState == STATE_HIGHSCORES)
        {
                glTranslatef(0.0f,0.0f,-5.0f);

                highscoreList.draw(gamefield.getScore());

                snowFlakes.update(frametime);                        
        }
        else if (gameState == STATE_CREDITS)
        {
                glTranslatef(0.0f,0.0f,-5.0f);

                credits.update(frametime);

                snowFlakes.update(frametime);         
        }
        //When inputting new highscore's name
        else if (gameState == STATE_INPUT)
        {
                glTranslatef(0.0f,0.0f,-5.0f);

                inputDialog.draw();

                //Draw text that shows what score and what place
                sprintf(buffer,"You've got a highscore. %d. place with %d points!",
                        highscoreList.isHighscore(gamefield.getScore()),
                        gamefield.getScore());

                font.printAt(0,0,buffer,true);

                snowFlakes.update(frametime);        
        }
        //No highscore has been done..
        else if (gameState == STATE_NOHIGHSCORE)
        {
                glTranslatef(0.0f,0.0f,-5.0f);

                sprintf(buffer, "Sorry no highscore with %d!", gamefield.getScore());
                //Calculate the exact middle in the X-Axis (yoooo)
                font.printAt(320-(strlen(buffer)*10)/2,250,buffer);
                
                snowFlakes.update(frametime);                           
        }
        
}

void SGLWindow::keyDown(int key)
{
        /** Any key is okay during end scene.
            We will simply quit
            **/

        if (gameState == STATE_END)
        {
                SDL_Event sent;
               	sent.type = SDL_QUIT;

                SDL_PushEvent(&sent);        
        }
        else if (gameState == STATE_INPUT)
        {
                //Make sure it's a range that our font will
                //support. Secondly these are the only important chars..
                if ( (key >= ' ' && key <= 'z') )
                {                     
                        inputDialog.keyEvent(key);        
                }
        }
        else if (gameState == STATE_PRESENTS || gameState == STATE_LOGO ||
                 gameState == STATE_INTRO || gameState == STATE_NOHIGHSCORE)
        {
                gameState = STATE_MENU;
                return;
        }
        
	switch(key)
 	{
    	case SDLK_ESCAPE:
                //only if in menu, we will quit.
                //Otherwise go back to menu
                if (gameState == STATE_MENU)
                {
                        gameState = STATE_END;
                        soundSystem.stopAll();
                        soundSystem.playSample(SAMPLE_MERRYCHRISTMAS);
                }
                else
                {
                        //if in options mode, go back to main menu
                        if (gameState == STATE_OPTIONS)
                                menu.back();
                        
                        gameState = STATE_MENU;
                }
         	break;
        case SDLK_LEFT:
                if (gameState == STATE_GAME)
                        gamefield.moveLeft();
                else if (gameState == STATE_OPTIONS)
                        menu.enter();
                break;
        case SDLK_RIGHT:
                if (gameState == STATE_GAME)
                        gamefield.moveRight();
                else if (gameState == STATE_OPTIONS)
                        menu.enter();
                     break;
        case SDLK_DOWN:
                if (gameState == STATE_GAME)
                        gamefield.moveDown();
                else if (gameState == STATE_MENU || gameState == STATE_OPTIONS)
                        menu.moveDown();
     
                break;
        case SDLK_UP:
                if (gameState == STATE_GAME)
                        gamefield.changeForm();
                else if (gameState == STATE_MENU || gameState == STATE_OPTIONS)
                        menu.moveUp();
                break;
        case SDLK_RETURN:
                if (gameState == STATE_MENU || gameState == STATE_OPTIONS)
                {
                        //Backup last settings for options
                        bool backupMusic = configFile->isMusic();
                        bool backupFullscreen = configFile->isFullscreen();

                        SGameState state = menu.enter();
                        
                        //If music option has changed, we will update music's
                        //state, play or stop it
                        if (backupMusic != configFile->isMusic())
                        {
                                if (configFile->isMusic())
                                        soundSystem.playSample(SAMPLE_BACKGROUND);
                                else
                                        soundSystem.stopAll();
                        }

                        //If fullscreen option has changed..
                        if (backupFullscreen != configFile->isFullscreen())
                        {
                                SDL_WM_ToggleFullScreen(mainSurface);
                        }
            
                        /** Catch all for other states
                         **/
                        if (state != SGameState(-1) )
                        {
                                gameState = state;
                                
                                /** Depending on state **/
                                if (gameState == STATE_GAME)
                                {
                                        gamefield.reset();
                                        gamefield.newBrick();

                                }
                                else if (gameState == STATE_CREDITS)
                                {
                                        credits.reset();
                                }
                        }
                 
                }
                //if return during input, add highscore
                //and go into highscore screem
                else if (gameState == STATE_INPUT)
                {
                        //add it.
                        highscoreList.addHighscore(inputDialog.getInput(),
                                                   gamefield.getScore());
                        //highscore.
                        gameState = STATE_HIGHSCORES;                                                                        
                }
                break;
        case SDLK_SPACE:
                if (gameState == STATE_GAME)
                {
                        gamefield.reset();
                        gamefield.newBrick();
                }
                else if (gameState == STATE_OPTIONS)
                {
                        menu.enter();
                }
                else if (gameState == STATE_INPUT)
                {
                        inputDialog.keyEvent(' ');        
                }
        case SDLK_BACKSPACE:
                //erase last char. Simply backspace.
                if (gameState == STATE_INPUT)
                {
                        inputDialog.backspace();
                }
                break;
        //Pause game
        case 'p':
                if (gameState == STATE_GAME)
                {
                        gamefield.pause();
                }
        }
}
